#include "../include/PCLFeatures.h"

//----------NormalEstimation-------------------
void NormalEstimation::computeNormal (pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, double radius_search)
{
  // Estimate the normals.
  pcl::NormalEstimation<pcl::PointXYZ, pcl::Normal> ne;
  ne.setInputCloud(cloud);

  pcl::search::KdTree<pcl::PointXYZ>::Ptr kdtree(new pcl::search::KdTree<pcl::PointXYZ>);
  ne.setSearchMethod(kdtree);
  ne.setRadiusSearch(radius_search);

  pcl::PointCloud<pcl::Normal>::Ptr normals (new pcl::PointCloud<pcl::Normal>);
  ne.compute(*normals);
  normals_.reset (new pcl::PointCloud<pcl::Normal>);
  *normals_ = *normals;
}

//-----------------VFH---------------
void Feature::computeVFH (pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, double radius_search)
{
  // Object for storing the VFH descriptors.
  pcl::PointCloud<pcl::VFHSignature308>::Ptr descriptors (new pcl::PointCloud<pcl::VFHSignature308>);

  // Estimate the normals.
  NormalEstimation ne;
  ne.computeNormal (cloud, radius_search);

  // VFH estimation object.
  pcl::VFHEstimation<pcl::PointXYZ, pcl::Normal, pcl::VFHSignature308> vfh;
  vfh.setInputCloud(cloud);
  vfh.setInputNormals(ne.normals_);

  pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ> () );
  vfh.setSearchMethod(tree);
  // Optionally, we can normalize the bins of the resulting histogram,
  // using the total number of points.
  vfh.setNormalizeBins(true);
  // Also, we can normalize the SDC with the maximum size found between
  // the centroid and any of the cluster's points.i
  vfh.setNormalizeDistance(false);
  vfh.compute(*descriptors);

  descriptors_vfh.reset(new pcl::PointCloud<pcl::VFHSignature308>);
  descriptors_vfh = descriptors;

  std::cerr << "computation of vfh is done" << std::endl;
}

//-----------------FPFH---------------
void Feature::computeFPFH (pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, double radius_normal, double radius_fpfh)
{
  // Object for storing the CVFH descriptors.
  pcl::PointCloud<pcl::FPFHSignature33>::Ptr descriptors (new pcl::PointCloud<pcl::FPFHSignature33>);

  // Estimate the normals.
  NormalEstimation ne;
  ne.computeNormal (cloud, radius_normal);

  // VFH estimation object.
  pcl::FPFHEstimation<pcl::PointXYZ, pcl::Normal, pcl::FPFHSignature33> fpfh;
  fpfh.setInputCloud(cloud);
  fpfh.setInputNormals(ne.normals_);

  pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ> () );
  fpfh.setSearchMethod(tree);

  // using the total number of points.
  fpfh.setRadiusSearch (radius_fpfh);

  fpfh.compute(*descriptors);

  descriptors_fpfh.reset(new pcl::PointCloud<pcl::FPFHSignature33>);
  descriptors_fpfh = descriptors;

  std::cerr << "computation of fpfh is done" << std::endl;
}


//-----------------OVFH---------------
void Feature::computeOVFH (pcl::PointCloud<pcl::PointXYZ>::Ptr &cloud, double radius_search)
{
  // Object for storing the VFH descriptors.
  pcl::PointCloud<pcl::VFHSignature308>::Ptr descriptors (new pcl::PointCloud<pcl::VFHSignature308>);

  // Estimate the normals.
  NormalEstimation ne;
  ne.computeNormal (cloud, radius_search);

  // VFH estimation object.
  pcl::VFHEstimation<pcl::PointXYZ, pcl::Normal, pcl::VFHSignature308> vfh;
  vfh.setInputCloud(cloud);
  vfh.setInputNormals(ne.normals_);

  pcl::search::KdTree<pcl::PointXYZ>::Ptr tree (new pcl::search::KdTree<pcl::PointXYZ> () );
  vfh.setSearchMethod(tree);
  // Optionally, we can normalize the bins of the resulting histogram,
  // using the total number of points.
  vfh.setNormalizeBins(true);
  // Also, we can normalize the SDC with the maximum size found between
  // the centroid and any of the cluster's points.i
  vfh.setNormalizeDistance(false);
  vfh.compute(*descriptors);

  /*-----------PCA----------------*/
  // Placeholder for the 3x3 covariance matrix at each surface patch
   Eigen::Matrix3f covariance_matrix;
   // 16-bytes aligned placeholder for the XYZ centroid of a surface patch
   Eigen::Vector4f xyz_centroid;

   // Estimate the XYZ centroid
   pcl::compute3DCentroid (*cloud, xyz_centroid);

   // Compute the 3x3 covariance matrix
   pcl::computeCovarianceMatrix (*cloud, xyz_centroid, covariance_matrix);

   std::cout<<covariance_matrix<<std::endl;

   Eigen::EigenSolver<Eigen::Matrix3f> es3d(covariance_matrix);

   Eigen::VectorXf evals; //we'll extract the eigenvalues to here

   evals= es3d.eigenvalues().real(); // grab just the real parts

   double min_lambda = evals[0]; //initialize the hunt for min eval
   Eigen::Vector3cf complex_vec; // here is a 3x1 vector of double-precision, complex numbers
   Eigen::Vector3f est_plane_normal;
   complex_vec=es3d.eigenvectors().col(0); // here's the first e-vec, corresponding to first e-val
   est_plane_normal = complex_vec.real();  //strip off the real part


   double lambda_test;
   int i_normal=0;
   //loop through "all" ("both", in this 3-D case) the rest of the solns, seeking min e-val
   for (int ivec=1;ivec<3;ivec++) {
       lambda_test = evals[ivec];
       if (lambda_test<min_lambda) {
       min_lambda =lambda_test;
               i_normal= ivec; //this index is closer to index of min eval
       est_plane_normal = es3d.eigenvectors().col(ivec).real();
       }
   }


   Eigen::Vector3f viewpoint (0, 0, 0);
   Eigen::Vector3f d_vp_p( viewpoint[0] - xyz_centroid[0], viewpoint[1] - xyz_centroid[1], viewpoint[2] - xyz_centroid[2]);
   d_vp_p.normalize ();

   if(est_plane_normal.dot(d_vp_p)>=0)
       est_plane_normal=est_plane_normal;
   else
       est_plane_normal=-est_plane_normal;

    std::cout<<"to VP"<<std::endl<<est_plane_normal<<std::endl;

    Eigen::Vector3f o_vp;
    o_vp = d_vp_p.cross(est_plane_normal);
    o_vp.normalize ();

    int size=0;
    size=cloud->points.size();
    double hist_incr;
    hist_incr = 100.0 / static_cast<double> (size);
    std::vector<float> hist_o_vp(128,0);
    for(int i=0;i<size;++i)
    {
        Eigen::Vector3f normal (ne.normals_->points[i].normal[0],
                                ne.normals_->points[i].normal[1],
                                ne.normals_->points[i].normal[2]);

        double alpha = (normal.dot (o_vp) + 1.0) * 0.5;
        int fi = static_cast<int> (floor (alpha * 128));
        if (fi < 0)
          fi = 0;
        if (fi > (127 ))
          fi = 127;
        // Bin into the histogram
        hist_o_vp [fi] += static_cast<float> (hist_incr);

    }
    for (int d = 0; d < hist_o_vp.size (); ++d)
      descriptors->points[0].histogram[d + 180] = hist_o_vp[d];


  descriptors_ovfh.reset(new pcl::PointCloud<pcl::VFHSignature308>);
  descriptors_ovfh = descriptors;

  std::cerr << "computation of ovfh is done" << std::endl;
}
